#include "GetObjects.js"

/*
  CutWalls 0.9
  12/30/2022, Larry Fahnoe, fahnoe@FahnoeTech.com
  https://moi3d.com/forum/messages.php?webtag=MOI&msg=10907.1

  Primarily intended for cutting windows and doors in the walls of
  architectural models. It can optionally create a solid frame and a
  surface to represent a pane of glass, either of which may be named.

  Windows to be cut are expected to be represented by closed, planar
  curves located on the wall's external surface. Window curves can be
  any shape. If glass is to be added it will be inset from the
  external surface of the wall.

  Walls may be made up of surfaces and/or solids. FindPts() seeks to
  find the thickness of the wall by looking for selected wall (base)
  objects on either side of the window curve. Probe depth (which
  defaults to 15 inches) defines how far on either side. If a frame is
  to be made it will be extruded to this thickness.
  
  Based on MoI's BooleanDifference command.
*/

function GetIntersectPts( normal, objs)
{
    // Set up an intersection between the base objects and the 2D
    // curve (normal).
    objs.addObject( normal);
    var factory = moi.command.createFactory( 'intersect');
    factory.setInput( 0, objs);
    var intersectObjs = factory.calculate();
    factory.cancel();
    
    // Extract points intersecting the normal
    var intersectPts = new Array;
    for ( var i = 0; i < intersectObjs.length; ++i)
    {
	var obj = intersectObjs.item( i);
        if ( obj.isPointObject)
            intersectPts.push( obj.pt);
    }

    // Since ObjectLists appear to be passed by reference don't leave
    // the normal curve on the objs list (as it would show up on the
    // baseObjs list)
    objs.removeObject( normal);

    // Return the array of intersecting points
    return intersectPts;
}

// Returns an array of 5 points along planar curve's normal:
//   points[ 0] = first side
//   points[ 1] = opposite side
//   points[ 2] = first side - .1 unit offset
//   points[ 3] = opposite side + .1 unit offset
//   points[ 4] = 10% inside of first side (for glass)
function FindPts( curve, objs, probeDepth)
{
    // Starting with a closed, planar curve, get the plane's
    // coordinate frame
    var cf = curve.planarFrame;
    var points = new Array;

    // Create a probe line normal to the curve, extending on both
    // sides of the curve
    var factory = moi.command.createFactory( 'line');
    factory.setInput( 0, cf.evaluate( 0, 0, -probeDepth));
    factory.setInput( 1, cf.evaluate( 0, 0, probeDepth));
    var normal = factory.calculate().item( 0);
    factory.cancel();

    // Find the min/max points (wall thickness) using the curve
    // parameter t
    var intersectPts = GetIntersectPts( normal, objs);
    if ( intersectPts.length == 0)
        return false;
    var min = 1;
    var max = 0;
    var minPt = null;
    var maxPt = null;
    for ( var i = 0; i < intersectPts.length; ++i)
    {
        var t = normal.dropPoint( intersectPts[ i]);
        if ( t <= min)
        {
            min = t;
            minPt = intersectPts[ i];
        }        
        if ( t >= max)
        {
            max = t;
            maxPt = intersectPts[ i];
        }        
    }

    // Make sure the minPt is on the side with the original curve
    var origin = normal.dropPoint( cf.origin);
    if ( Math.abs( min - origin) > Math.abs( max - origin))
    {
        var x = min;
        var xPt = minPt;
        min = max;
        minPt = maxPt;
        max = x;
        maxPt = xPt;
    }
    points[ 0] = minPt;
    points[ 1] = maxPt;

    // Calculate points 2 & 3 10% outside of the wall for cutter, and
    // point 4 10% inside for glass
    var delta = max - min;
    points[ 2] = normal.evaluatePoint( min - 0.1 * delta);
    points[ 3] = normal.evaluatePoint( max + 0.1 * delta);
    points[ 4] = normal.evaluatePoint( min + 0.1 * delta);
    
    // Return the array of points
    return points;
}

function MakeFrame( curve, width, minPt, maxPt, name)
{
    var curves = moi.geometryDatabase.createObjectList();
    curves.addObject( curve);

    // If the curve isn't on the wall, move it onto the wall...
    var cf = curve.planarFrame;
    var origin = cf.origin;
    if ( !moi.vectorMath.pointsAreEqual( origin, minPt))
    {
        var factory = moi.command.createFactory( 'move');
        factory.setInput( 0, curves);
        factory.setInput( 1, origin);
        factory.setInput( 2, minPt);
        curves = factory.calculate();
        factory.cancel();
    }

    // Get outside point and create CF for use by offset
    var outsidePt = curves.getHighAccuracyBoundingBox().max;
    var outsideCf = moi.vectorMath.createFrame( outsidePt);

    // Create a curve offset to the outside
    var factory = moi.command.createFactory( 'offset');
    factory.setInput( 0, curves);
    factory.setInput( 1, width);
    factory.setInput( 2, outsideCf);
    factory.update();
    var outsideCrv = factory.calculate().item( 0);
    factory.cancel();

    // Extrude a solid frame from the 2 curves
    curves.addObject( outsideCrv);
    var factory = moi.command.createFactory( 'extrude');
    factory.setInput( 0, curves);
    factory.setInput( 1, maxPt);
    factory.setInput( 5, true); // Cap ends
    factory.update();
    var newObjs = factory.getCreatedObjects();
    factory.commit();

    // Name the newly created frame objects
    if ( name != null)
        newObjs.setProperty( 'name', name);

    // Return outside curve for use as a cutting object
    return outsideCrv;
}

function MakeGlass( curve, glassPt, name)
{
    var curves = moi.geometryDatabase.createObjectList();
    curves.addObject( curve);

    // Move the glass curve a little bit inside of the wall
    var cf = curve.planarFrame;
    var origin = cf.origin;
    var factory = moi.command.createFactory( 'move');
    factory.setInput( 0, curves);
    factory.setInput( 1, origin);
    factory.setInput( 2, glassPt);
    curves = factory.calculate();
    factory.cancel();

    var factory = moi.command.createFactory( 'planarsrf');
    factory.setInput( 0, curves);
    factory.update();
    var newObjs = factory.getCreatedObjects();
    factory.commit();

    // Name the newly created glass objects
    if ( name != null)
        newObjs.setProperty( 'name', name);
}

// Create wall cutter, optionally make frames and glass
function PrepWindow( curve, objs, probeDepth, makeFrame, width, frameName, makeGlass, glassName)
{
    var points = FindPts( curve, objs, probeDepth);
    if ( !points)
        return false;
    var frameMinPt = points[ 0];
    var frameMaxPt = points[ 1];
    var cutterMinPt = points[ 2];
    var cutterMaxPt = points[ 3];
    var glassPt = points[ 4];

    // Cutting curve is window curve unless making a frame
    var cuttingCurves = moi.geometryDatabase.createObjectList();
    var cuttingCurve = curve;
    if ( makeFrame)
        cuttingCurve = MakeFrame( curve, width, frameMinPt, frameMaxPt, frameName);
    cuttingCurves.addObject( cuttingCurve);

    // Move cutting curve a little beyond (outside of) the wall to
    // allow surfaces to be cut
    var factory = moi.command.createFactory( 'move');
    factory.setInput( 0, cuttingCurves);
    factory.setInput( 1, frameMinPt);
    factory.setInput( 2, cutterMinPt);
    cuttingCurves = factory.calculate();
    factory.cancel();

    // Extrude a solid cutter from the cutting curve
    var factory = moi.command.createFactory( 'extrude');
    factory.setInput( 0, cuttingCurves);
    factory.setInput( 1, cutterMaxPt);
    factory.setInput( 5, true); // Cap ends
    var cutter = factory.calculate().item( 0);
    factory.cancel();

    if ( makeGlass)
        MakeGlass( curve, glassPt, glassName);

    return cutter;
}

function doCutWalls()
{
    // First pick base objects to be cut (walls)
    var objectpicker = moi.ui.createObjectPicker();
    objectpicker.allowBReps();
    objectpicker.allowFaces();
    
    if ( !GetObjects( objectpicker))
	return;

    var baseObjs = objectpicker.objects;

    moi.ui.beginUIUpdate();
    moi.ui.hideUI( 'FirstSelectPrompt');
    moi.ui.showUI( 'SecondSelectPrompt');
    moi.ui.showUI( 'SecondPromptOptions');
    moi.ui.endUIUpdate();

    // Next pick curves representing openings (windows) to be cut
    objectpicker.resetAllow();
    objectpicker.allowCurves();

    if ( !GetObjects( objectpicker))
	return;
    
    var curves = objectpicker.objects;
    
    // Get probe depth
    var probeDepth = moi.ui.commandUI.probeDepth.value;
    if ( probeDepth <= 0)
    {
        moi.ui.alert( 'Please enter a positive probe depth');
        return;
    }

    // Get frame options
    var makeFrame = moi.ui.commandUI.makeFrame.value;
    var width = 0;
    var frameName = null;
    if ( makeFrame)
    {
        width = moi.ui.commandUI.width.value;
        frameName = moi.ui.commandUI.frameName.value;
        if ( width <= 0)
        {
            moi.ui.alert( 'Please enter a positive width');
            return;
        }
    }

    // Get glass options
    var makeGlass = moi.ui.commandUI.makeGlass.value;
    var glassName = null;
    if ( makeGlass)
    {
        glassName = moi.ui.commandUI.glassName.value;
    }

    // Allowing multiple windows and walls to be cut, and allowing
    // walls to be of differing thickness, loop through the planar
    // curves and prep windows one at a time. Save the resulting
    // cutters for a single boolean difference at the end (otherwise
    // the wall objects get duplicated for each window cut).
    var cutters = moi.geometryDatabase.createObjectList();
    for ( var i = 0; i < curves.length; ++i)
    {
        var curve = curves.item( i);
        if ( !curve.isClosed || !curve.isPlanar)
        {
            moi.ui.alert( 'Cutting curves need to be both planar and closed');
            return;
        }
        var cutter = PrepWindow( curve, baseObjs, probeDepth, makeFrame, width, frameName, makeGlass, glassName);
        if ( cutter)
            cutters.addObject( cutter);
    }

    // From Michael: In your script the thing that is different
    // compared to the regular boolean difference command is that the
    // cutting objects are "loose objects" generated inside the script
    // and haven't been added into the geometry database and haven't
    // ever been selected. It happens in that case that it gets the
    // default selection order value of 0 and that makes the "assign
    // top level object name from earliest selected original object"
    // function think that the cutting objects have the priority.

    // The result of the above is that top level solid object names
    // are deleted. Oddly, surface objects don't loose their names.
    
    cutters.setProperty( 'selected', true);

    // Boolean difference doesn't cut surfaces if the cutting object
    // is flush with the surfaces. Either the cutting object needs to
    // extend through the surfaces, or a use a trim operation.
    // PrepWindow() creates a cutter that extends through the walls.

    // Cut all the openings at once
    var factory = moi.command.createFactory( 'booleandifference');
    factory.setInput( 0, baseObjs);
    factory.setInput( 1, cutters);
    factory.setInput( 2, false); // Don't keep objects 
    factory.commit();

    // Conditionally delete window curves
    if ( moi.ui.commandUI.keepObjects.value == false)
        moi.geometryDatabase.removeObjects( curves);
}

doCutWalls();
